##
##  OpenPKG Tool Chain
##  Copyright (c) 2000-2003 Cable & Wireless Deutschland GmbH
##  Copyright (c) 2000-2003 The OpenPKG Project <http://www.openpkg.org/>
##  Copyright (c) 2000-2003 Ralf S. Engelschall <rse@engelschall.com>
##
##  Permission to use, copy, modify, and distribute this software for
##  any purpose with or without fee is hereby granted, provided that
##  the above copyright notice and this permission notice appear in all
##  copies.
##
##  THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
##  WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
##  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
##  IN NO EVENT SHALL THE AUTHORS AND COPYRIGHT HOLDERS AND THEIR
##  CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
##  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
##  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
##  USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
##  ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
##  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
##  OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
##  SUCH DAMAGE.
##
##  api/OpenPKG/Ctx.pm: API Context Class
##

package OpenPKG::Ctx;

use OpenPKG::Obj;
use Carp;
use vars qw($AUTOLOAD);

@ISA       = qw(OpenPKG::Obj);
@EXPORT_OK = qw();

sub new ($) {
    my $proto = shift;
    my $class = ref($proto) || $proto;
    my $self = {};
    bless ($self, $class);

    $self->{'prefix'}   = ($ENV{'OPENPKG_PREFIX'}         || '/openpkg');
    $self->{'cmdprog'}  = ($ENV{'OPENPKG_TOOLS_CMDPROG'}  || '/openpkg/bin/openpkg');
    $self->{'cmdname'}  = ($ENV{'OPENPKG_TOOLS_CMDNAME'}  || '');
    $self->{'cmdstack'} = ($ENV{'OPENPKG_TOOLS_CMDSTACK'} || '');
    $self->{'perl'}     = $^X;

    return $self;
}

#   get and set object attributes
sub AUTOLOAD {
    my $self = shift;
    my $type = ref($self) or croak "$self is not an object";

    #   determine name of requested method
    my $name = $AUTOLOAD;
    $name =~ s/.*://;

    #   make sure the corresponding attribute exists
    if (not defined($self->{$name})) {
        croak "attribute `$name' not exists in class `$type'";
    }

    #   get or set attribute
    my $val = $self->{$name};
    if (@_) {
        $self->{$name} = shift;
    }
    return $val;
}

1;

__END__

=pod

=head1 NAME

B<OpenPKG::Ctx> - Context Class

=head1 DESCRIPTION

B<OpenPKG::Ctx> is the Context class of the OpenPKG Tool Chain Perl API.
It provides access to the run-time context which the B<openpkg> execution
frontend usually has provided initially.

=head1 METHODS

=over 4

=item C<$ctx = >B<new>C< OpenPKG::Ctx;>

=item C<$ctx-E<gt>>B<destroy>C<;>

Create and destroy object.

=item C<$path = $ctx-E<gt>>B<tool>C<;>

=item C<$ctx-E<gt>>B<tool>C<($path);>

Get and set path to the B<openpkg> execution frontend. This
is preinitialized to the value of environment variable
C<$OPENPKG_TOOLS_TOOL> as provided by the B<openpkg> execution frontend
initially.

=item C<$path = $ctx-E<gt>>B<home>C<;>

=item C<$ctx-E<gt>>B<home>C<($path);>

Get and set path to OpenPKG Tool Chain home directory. This
is preinitialized to the value of environment variable
C<$OPENPKG_TOOLS_HOME> as provided by the B<openpkg> execution
frontend initially.

=item C<$path = $ctx-E<gt>>B<instance>C<;>

=item C<$ctx-E<gt>>B<instance>C<($path);>

Get and set path to OpenPKG instance (prefix directory).
This is preinitialized to the value of environment variable
C<$OPENPKG_TOOLS_INST> as provided by the B<openpkg> execution frontend
initially if an OpenPKG instance was found.

=item C<$path = $ctx-E<gt>>B<perl>C<;>

=item C<$ctx-E<gt>>B<perl>C<($path);>

Get and set path to Perl interpreter executable. This
is preinitialized to the value of environment variable
C<$OPENPKG_TOOLS_PERL> as provided by the B<openpkg> execution
frontend initially.

=head1 SEE ALSO

L<OpenPKG(3)>.

=head1 AUTHOR

The OpenPKG Project E<lt>http://www.openpkg.org/E<gt>

=cut

